@@ -15,7 +15,6 @@
 
 #include "list.h"
 #include "translate.h"
-#include "zstd-decompress.h"
 
 #define MAX(x, y) ((x) > (y) ? (x) : (y))
 
@@ -47,6 +46,12 @@ struct tftp_client {
 	size_t wsize;
 	unsigned int timeoutms;
 	off_t seek;
+
+	uint8_t *blk_buf;
+	uint8_t *rw_buf;
+	size_t rw_buf_size;
+	size_t blk_offset;
+	uint16_t blk_expected;
 };
 
 static struct list_head readers = LIST_INIT(readers);
@@ -57,11 +62,8 @@ static ssize_t tftp_send_data(struct tftp_client *client,
 {
 	ssize_t len;
 	size_t send_len;
-	char *buf;
-	char *p;
-
-	buf = malloc(4 + client->blksize);
-	p = buf;
+	char *buf = client->blk_buf;
+	char *p = buf;
 
 	*p++ = 0;
 	*p++ = OP_DATA;
@@ -92,11 +94,7 @@ static ssize_t tftp_send_data(struct tftp_client *client,
 	}
 
 	// printf("[TQFTP] Sending %zd bytes of DATA\n", send_len);
-	len = send(client->sock, buf, send_len, 0);
-
-	free(buf);
-
-	return len;
+	return send(client->sock, buf, send_len, 0);
 }
 
 
@@ -252,7 +250,7 @@ static void handle_rrq(const char *buf, size_t len, struct sockaddr_qrtr *sq)
 	size_t blksize = 512;
 	unsigned int timeoutms = 1000;
 	size_t rsize = 0;
-	size_t wsize = 0;
+	size_t wsize = 1;
 	off_t seek = 0;
 	bool do_oack = false;
 	int sock;
@@ -316,6 +314,19 @@ static void handle_rrq(const char *buf, size_t len, struct sockaddr_qrtr *sq)
 	client->wsize = wsize;
 	client->timeoutms = timeoutms;
 	client->seek = seek;
+	client->rw_buf_size = blksize * wsize;
+
+	client->blk_buf = calloc(1, blksize + 4);
+	if (!client->blk_buf) {
+		printf("[TQFTP] Memory allocation failure\n");
+		return;
+	}
+
+	client->rw_buf = calloc(1, client->rw_buf_size);
+	if (!client->rw_buf) {
+		printf("[TQFTP] Memory allocation failure\n");
+		return;
+	}
 
 	// printf("[TQFTP] new reader added\n");
 
@@ -343,7 +354,7 @@ static void handle_wrq(const char *buf, size_t len, struct sockaddr_qrtr *sq)
 	size_t blksize = 512;
 	unsigned int timeoutms = 1000;
 	size_t rsize = 0;
-	size_t wsize = 0;
+	size_t wsize = 1;
 	off_t seek = 0;
 	bool do_oack = false;
 	int sock;
@@ -398,6 +409,20 @@ static void handle_wrq(const char *buf, size_t len, struct sockaddr_qrtr *sq)
 	client->wsize = wsize;
 	client->timeoutms = timeoutms;
 	client->seek = seek;
+	client->rw_buf_size = blksize * wsize;
+	client->blk_expected = 1;
+
+	client->blk_buf = calloc(1, blksize + 4);
+	if (!client->blk_buf) {
+		printf("[TQFTP] Memory allocation failure\n");
+		return;
+	}
+
+	client->rw_buf = calloc(1, client->rw_buf_size);
+	if (!client->rw_buf) {
+		printf("[TQFTP] Memory allocation failure\n");
+		return;
+	}
 
 	// printf("[TQFTP] new writer added\n");
 
@@ -494,14 +519,14 @@ static int handle_writer(struct tftp_client *client)
 	struct sockaddr_qrtr sq;
 	uint16_t block;
 	size_t payload;
-	char buf[516];
+	uint8_t *buf = client->blk_buf;
 	socklen_t sl;
 	ssize_t len;
 	int opcode;
 	int ret;
 
 	sl = sizeof(sq);
-	len = recvfrom(client->sock, buf, sizeof(buf), 0, (void *)&sq, &sl);
+	len = recvfrom(client->sock, buf, client->blksize + 4, 0, (void *)&sq, &sl);
 	if (len < 0) {
 		ret = -errno;
 		if (ret != -ENETRESET)
@@ -523,17 +548,48 @@ static int handle_writer(struct tftp_client *client)
 	}
 
 	payload = len - 4;
+	buf += 4;
+
+	/* Check if we recieved expected block */
+	if (block != client->blk_expected) {
+		uint16_t blk_expected = client->blk_expected;
+
+		printf("[TQFTP] Block number out of sequence: %d (expected %d)\n",
+			 block, blk_expected);
+		tftp_send_error(client->sock, 4, "Block number out of sequence");
+
+		/* Set blk_expected to beginning of current window */
+		if ((blk_expected % client->wsize) == 0)
+			blk_expected -= client->wsize + 1;
+		else
+			blk_expected -= (blk_expected % client->wsize) - 1;
+
+		client->blk_expected = blk_expected;
+		client->blk_offset = 0;
 
-	ret = write(client->fd, buf + 4, payload);
-	if (ret < 0) {
-		/* XXX: report error */
-		printf("[TQFTP] failed to write data\n");
 		return -1;
 	}
 
-	tftp_send_ack(client->sock, block);
+	client->blk_expected++;
 
-	return payload == 512 ? 1 : 0;
+	/* Copy the data to the destination buffer */
+	memcpy(client->rw_buf + client->blk_offset, buf, payload);
+	client->blk_offset += payload;
+
+	/* Write to file if all the wsize blocks are recieved */
+	if (block % client->wsize == 0) {
+		ret = write(client->fd, client->rw_buf, client->blk_offset);
+		if (ret < 0) {
+			/* XXX: report error */
+			printf("[TQFTP] failed to write data\n");
+			return -1;
+		}
+
+		client->blk_offset = 0;
+		tftp_send_ack(client->sock, block);
+	}
+
+	return payload == client->blksize ? 1 : 0;
 }
 
 static void client_close_and_free(struct tftp_client *client)
@@ -541,6 +597,8 @@ static void client_close_and_free(struct tftp_client *client)
 	list_del(&client->node);
 	close(client->sock);
 	close(client->fd);
+	free (client->blk_buf);
+	free (client->rw_buf);
 	free(client);
 }
 
@@ -571,8 +629,6 @@ int main(int argc, char **argv)
 		exit(1);
 	}
 
-	zstd_init();
-
 	for (;;) {
 		FD_ZERO(&rfds);
 		FD_SET(fd, &rfds);
@@ -674,7 +730,6 @@ int main(int argc, char **argv)
 	}
 
 	close(fd);
-	zstd_free();
 
 	return 0;
 }
