@@ -24,6 +24,7 @@
 #ifndef ANDROID
 #define FIRMWARE_BASE	"/lib/firmware/"
 #define TQFTPSERV_TMP	"/tmp/tqftpserv"
+#define UPDATES_DIR	"updates/"
 #else
 #define FIRMWARE_BASE	"/vendor/firmware/"
 #define TQFTPSERV_TMP	"/data/vendor/tmp/tqftpserv"
@@ -129,17 +130,25 @@ static int translate_readonly(const char *file)
 		}
 
 		/* now try with base path */
-		if (strlen(FIRMWARE_BASE) + strlen(firmware_value) + 1 +
+		if (strlen(FIRMWARE_BASE) + strlen(UPDATES_DIR) + strlen(firmware_value) + 1 +
 		    strlen(file) + 1 > sizeof(path))
 			continue;
 
 		strcpy(path, FIRMWARE_BASE);
+		strcat(path, UPDATES_DIR);
 		strcat(path, firmware_path);
 		strcat(path, "/");
 		strcat(path, file);
 
 		fd = open_maybe_compressed(path);
+		if (fd < 0) {
+			strcpy(path, FIRMWARE_BASE);
+			strcat(path, firmware_path);
+			strcat(path, "/");
+			strcat(path, file);
 
+			fd = open_maybe_compressed(path);
+		}
 		if (fd >= 0)
 			break;
 
@@ -217,16 +226,18 @@ int translate_open(const char *path, int flags)
  */
 static int open_maybe_compressed(const char *path)
 {
+	char *path_with_zstd_extension = NULL;
 	int fd = -1;
 
-	if (access(path, F_OK) == 0) {
-		fd = open(path, O_RDONLY);
-	} else {
-		char *path_with_zstd_extension = NULL;
-		asprintf(&path_with_zstd_extension, "%s%s", path, ZSTD_EXTENSION);
+	if (access(path, F_OK) == 0)
+		return open(path, O_RDONLY);
+
+	asprintf(&path_with_zstd_extension, "%s%s", path, ZSTD_EXTENSION);
+
+	if (access(path_with_zstd_extension, F_OK) == 0)
 		fd = zstd_decompress_file(path_with_zstd_extension);
-		free(path_with_zstd_extension);
-	}
+
+	free(path_with_zstd_extension);
 
 	return fd;
 }
