@@ -8,8 +8,14 @@
 
 #include <stdbool.h>
 
-void zstd_init();
-void zstd_free();
+#ifdef HAVE_ZSTD
 int zstd_decompress_file(const char *filename);
+#else
+static int zstd_decompress_file(const char *filename)
+{
+	fprintf(stderr, "Built without ZSTD support\n");
+	return -1;
+}
+#endif
 
 #endif
